/* C base neural network and derived back
 propagation neural network based on the equations in
 Rummelhart and McClelland'
 Processing and the shareware back-propagation neural
 network written in C by Eberhart and Dobbins and
 published in Mfcr0 c0r-Mcopid (Jan./Feb. l990). */

/* Includes */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <time.h>
#include "vector.h"

/* Constantes */
#define ESC 27
#define ITEMS 8
#define ERRORLEVEL 0.02

/* Definicion de las matrices de almacenamiento dinamico */
MATRIX  Out0;                /* Capa de entrada */
MATRIX  Out1;                /* Capa oculta */
MATRIX  Delta1;              /* Matriz de deltas de la capa oculta */
MATRIX  Delw1;               /* Cambio de pesos de entrada:oculta */
MATRIX  W1;                  /* Pesos de entrada:oculta */
MATRIX  Out2;                /* Capa de salida */
MATRIX  Outend;              /* Capa de salida final */
MATRIX  Delta2;              /* Matriz de deltas de la capa de salida */
MATRIX  Delw2;               /* Cambio de los pesos de oculta:salida */
MATRIX  W2;                  /* Pesos de oculta:salida */
MATRIX  target;              /* Salida deseada */
VECTOR  Context;	     /* Unidades de contexto */

/*--------------------------* PROGRAMA PRINCIPAL *--------------------------*/

void main(int argc, char *argv[])
{

float   Eta = 0.15,          /* Parmetro de aprendizaje por defecto */
Alpha = 0.075;               /* Momentum por defecto */
long     nReportErrors = 1;  /* Informe de error cada nReportErrors */
float   ErrorLevel ;         /* Nivel de error  */
char    MonitorError =0 ;    /* True cuando es presionada una tecla */
double Error;               /* Suma al cuadrado del error */
register long H;              /* Indice de la capa oculta */
register long I;              /* Indice de la capa de entrada */
register long J;              /* Indice de la capa de salida */

long nInputNodes,             /* N de unidades de entrada */
nHiddenNodes,                /* N de unidades de la capa oculta */
nOutputNodes,                /* N de unidades de salida */
nIterations,                 /* N de iteraciones */
nPatterns,                   /* N de patrones */
nLetters,		     /* N de letras por patrn */
MaxLetters,		     /* N mximo de letras que puede tener una patrn */
nRuns,                       /* N de ejecuciones */
P,                           /* Indice del nmero de patrn */
Q,                           /* Indice del nmero de iteraciones */
R,                           /* Indice del nmero de ejecuciones */
L,			     /* Indice para el nmero de letras por patrn */
C;			     /* Indice para el nmero de capas de contexto */

FILE      *fpRun,            /* Fichero de ejecucin  */
	  *fpPattern,        /* Fichero de patrones de entrada */
	  *fpWeights,        /* Fichero de pesos iniciales */
	  *fpWeightsOut,     /* Fichero de pesos finales */
	  *fpResults,        /* Fichero de resultados */
	  *fpError;          /* Fichero de error */
char  szResults[66];         /* Nombre de varios ficheros */
char  szError[66];
char  szPattern[66];
char  szWeights[66];
char  szWeightsOut[66];
char  *programe = * argv;

char pattern[100];	     /* Un patrn de entrada */

int c;			     /* Tecla pulsada por el usuario */
double Sum, Dw;

/*____________ Lectura de los argumentos desde la lnea de comandos________ */

for (; argc > 1; argc--)
{
	char *arg = *++argv;
	if (*arg != '-')
		 break;
	switch (*++arg)
	{
		case 'e':   sscanf(++arg,  "%ld",  &nReportErrors);  break;
		case 'd':   sscanf(++arg,  "%f",  &ErrorLevel); break;
		default:    break;
	}
}

if (argc < 2)
{
	fprintf(stderr, "Usage: %s {-en -df} runfilename\n",programe);
	fprintf(stderr, " -en => report error every n iterations\n");
	fprintf(stderr, " -df => done if sum squared error < f\n");
	exit(1);
}

/* Abrimos el fichero de ejecuciones */

if ((fpRun = fopen(*argv, "r")) == NULL)
{
	fprintf(stderr,"%s: Can't open file %s \n",programe,*argv);
	exit(1);
}

fscanf(fpRun, "%ld",&nRuns); /* Leemos el nmero de ejecuciones */

/*________________ Lectura de los diferentes datos de entrada ______________*/

clrscr();  /* Limpia la pantalla */

fscanf(fpRun, "%f",&ErrorLevel); /* Lectura del nivel de error */

for (R= 0; R< nRuns; R++)
{

/* Leemos las caractersticas de una ejecucin y la informacin sobre la red */
fscanf(fpRun, "%s %s %s %s %s %ld %ld %ld %ld %ld %ld %f %f",
	 szResults,                /* Fichero de resultados */
	 szError,                  /* Fichero de error */
	 szPattern,                /* Fichero de patrones */
	 szWeights,                /* Fichero de pesos iniciales */
	 szWeightsOut,             /* Fichero de pesos finales */
	 &nPatterns,               /* N de patrones a leer */
	 &MaxLetters,              /* N mximo de letras permitidas en un patrn */
	 &nIterations,             /* N de iteraciones sobre el cjto. de patrones */
	 &nInputNodes,             /* N de unidades de entrada */
	 &nHiddenNodes,            /* N de unidades de la capa oculta */
	 &nOutputNodes,            /* N de unidades de salida */
	 &Eta,                     /* Parmetro de aprendizaje */
	 &Alpha);                  /* Factor de momentum */

printf("\n\nEjecucin en Batch %ld:\n", R+1);
printf("\tN de Patrones    = %ld\n", nPatterns);
printf("\tN de Iteraciones = %ld\n", nIterations);
printf("\tCota de error     = %f\n", ErrorLevel);
printf("\tParmetro de aprendizaje = %f\n", Eta);
printf("\tParmetro momentum       = %f\n", Alpha);
printf("\tN de unidades en la capa de entrada  = %ld\n", nInputNodes);
printf("\tN de unidades de contexto            = %ld\n", nHiddenNodes);
printf("\tN de unidades en la capa oculta      = %ld\n", nHiddenNodes);
printf("\tN de unidades en la capa de salida   = %ld\n", nOutputNodes);


/* Ubicacin del almacenamiento dinmico apa los patrones de entrada y los nodos de la red */

AllocateMatrix(&Out0,  MaxLetters + 2,   nInputNodes);
AllocateMatrix(&Out1,  MaxLetters + 2,   nHiddenNodes);
AllocateMatrix(&Out2,  MaxLetters + 2,   nOutputNodes);
AllocateMatrix(&Outend,MaxLetters + 2,   nOutputNodes);
AllocateMatrix(&Delta2,MaxLetters + 2,   nOutputNodes);
AllocateMatrix(&Delta1,MaxLetters + 2,   nHiddenNodes);
AllocateMatrix(&target,MaxLetters + 2,   nOutputNodes);
AllocateMatrix(&Delw2, nOutputNodes,  nHiddenNodes + 1);
AllocateMatrix(&W2,    nOutputNodes,  nHiddenNodes + 1);
AllocateMatrix(&Delw1, nHiddenNodes,  nInputNodes + 1);
AllocateMatrix(&W1,    nHiddenNodes,  nInputNodes + 1);
AllocateVector(&Context, nHiddenNodes);

/*  Lectura de los pesos iniciales desde un fichero de entrada */
if ((fpWeights = fopen(szWeights,"r")) == NULL)
  {
  fprintf(stderr," %s: Canot open file %s \n",programe,szWeights);
  exit(1);
  } /* fin if */

/* Lectura de los pesos capa entrada:oculta  */
for (H = 0;  H < nHiddenNodes;  H++)
  for (I = 0; I <= nInputNodes; I++)
    {
    fscanf(fpWeights, "%lf", &W1[H][I]);
    Delw1[H][I] = 0.0;
    } /* fin for I */

/* Lectura de los pesos capa oculta:salida  */
for (J = 0;  J < nOutputNodes;  J++)
  for (H = 0; H <= nHiddenNodes; H++)
    {
    fscanf(fpWeights, "%lf", &W2[J][H]);
    Delw2[J][H] = 0.0;
    } /* fin for H */

/* Cerramos el fichero de pesos */
fclose(fpWeights);

/* Inicializa unidades de contexto */
for (C = 0;  C < nHiddenNodes;  C++)
  Context[H] = 0;

 /* Abrimos el fichero de salida para el error. */
if ((fpError = fopen(szError, "w")) == NULL)
  {
  fprintf(stderr," %s:  Can't open file %s \n",programe,szError);
  exit(1);
  } /* fin if */
fprintf(stderr, nIterations > 1 ? "Entrenando...\n" : "Testeando\n");

/* Abrimos el fichero final de activacin */
if ((fpResults = fopen(szResults,"w")) == NULL)
 {
 fprintf(stderr," %s:  Can't write file %s \n",programe,szResults);
 fpResults = stderr;
 }

/*__________________________ BUCLE DE PRINCIPAL ____________________________*/

for (Q = 0; Q < nIterations; Q++)
  {

  /* Abrimos el fichero de patrones */
  if ((fpPattern = fopen(szPattern, "r")) == NULL)
    {
    fprintf(stderr," %s: Can't Open file %s \n",programe,szPattern);
    exit(1);
    } /* fin if */

  for ( P = 0; P < nPatterns; ++P)
    {
    if (fscanf(fpPattern, "%s", pattern) == EOF)
       {
       fcloseall();
       printf(" No hay suficientes patrones !! (slo hay %ld)", P);
       exit(1);
       }

    nLetters = strlen(pattern);	/* N de letras del patrn */

    for (L = 0; L < nLetters - 1; L++)
      {

      /* Unidades de Entrada */
      for (I = 0; I < nInputNodes; I++)
	Out0[L][I] = 0;
      switch (pattern[L])
	  {
	  case 'B': Out0[L][0] = 1;
		    break;
	  case 'P': Out0[L][1] = 1;
		    break;
	  case 'S': Out0[L][2] = 1;
		    break;
	  case 'T': Out0[L][3] = 1;
		    break;
	  case 'V': Out0[L][4] = 1;
		    break;
	  case 'X': Out0[L][5] = 1;
		    break;
	  } /* fin switch */

      /* Unidades de la salida deseada */
      for (J = 0; J < nOutputNodes; J++)
	target[L][J] = 0;
      switch (pattern[L+1])
	{
	case 'E': target[L][0] = 1;
		  break;
	case 'P': target[L][1] = 1;
		  break;
	case 'S': target[L][2] = 1;
		  break;
	case 'T': target[L][3] = 1;
		  break;
	case 'V': target[L][4] = 1;
		  break;
	case 'X': target[L][5] = 1;
		  break;
	} /* fin switch */
      } /* fin for L */


/*_____________ Iteraciones para entrenear o testear la red ________________*/

    for (L = 0; L < nLetters - 1; L++)
      {

      /*           Proceso de Propagacin hacia delante               */

      /* Calculamos la salida de las neuronas de las capa oculta */
      for (H = 0; H < nHiddenNodes; H++)
	{
	Sum = W1[H][nInputNodes];           /* Aadimos el Bias */
	for (C = 0; C < nHiddenNodes; C++)  /* Aadimos la de contexto */
	    Sum += 0.5 * Context[C];
	for (I = 0; I < nInputNodes; I++)
	  Sum   +=   W1[H][I]  *  Out0[L][I];
	/* Calculo de la salida mediante una funcin de transferencia Sigmoide */
	Out1[L][H] = 1.0 / (1.0 + exp(-Sum));
	} /* fin del for H */

      /* Repetimos el proceso con las neuronas de la capa de salida */
      for (J = 0; J < nOutputNodes; J++)
	{
	Sum = W2[J][nHiddenNodes];
	for (H = 0; H < nHiddenNodes; H++)
	  Sum += W2[J][H] * Out1[L][H];
	/* Calculo de la salida con la funcin Sigmoide */
	Out2[L][J] = 1.0 / (1.0 + exp(-Sum));
	} /* fin del for J */

      /*           Proceso de retro_propagacin del error               */

      /* Calculamos los deltas para cada unidad de salida y capa patron de entrada */
      for (J = 0;  J < nOutputNodes;  J++)
	Delta2[L][J] = (target[L][J] - Out2[L][J]) *
		       Out2[L][J] * (1.0 - Out2[L][J]);

      /* Calculamos los deltas para cada neurona de la capa oculta y cada patrn de entrada */
      for (H = 0; H < nHiddenNodes; H++)
	{
	Sum = 0;
	for (J = 0; J < nOutputNodes; J++)
	  Sum += Delta2[L][J] * W2[J][H];
	/* Calculamos la salida usando la funcin Sigmoide */
	Delta1[L][H] = Sum * Out1[L][H] * (1.0 - Out1[L][H]);
	} /* fin for H */

      /*           Proceso de modificacin de los pesos de la red      */

      /* Adaptamos los pesos que unen la capa oculta con la salida */
      for (J = 0; J < nOutputNodes; J++)
	{
	Sum  =  Delta2[L][J];
	/* Calculamos el nuevo peso Bias para cada neurona de la capa de salida */
	Dw = Eta * Sum + Alpha * Delw2[J][nHiddenNodes];
	W2[J][nHiddenNodes] += Dw;
	Delw2[J][nHiddenNodes] = Dw; /* Delta para el Bias */

	/* Calculamos lo nuevos pesos */
	for (H = 0; H < nHiddenNodes; H++)
	  {
	  Sum  =  Delta2[L][J] * Out1[L][H];
	  Dw  =  Eta * Sum  +  Alpha * Delw2[J][H];
	  W2[J][H]    +=  Dw;
	  Delw2[J][H]  =  Dw;
	  } /* fin for H */
	} /* fin for J */

      /* Adaptamos los pesos que unen la capa de entrada con la oculta */
      for (H = 0; H < nHiddenNodes; H++)
	{
	Sum  =  Delta1[L][H];
	/* Calcumos el nuevo peso Bias para cada neurona de la capa oculta. */
	Dw = Eta * Sum + Alpha * Delw1[H][nInputNodes];
	W1[H][nInputNodes]   += Dw;
	Delw1[H][nInputNodes] = Dw;

	/* Calculamos los nuevos pesos */
	for (I = 0; I < nInputNodes; I++)
	  {
	  Sum = Delta1[L][H] * Out0[L][I];
	  Dw = Eta * Sum + Alpha * Delw1[H][I];
	  W1[H][I] += Dw;
	  Delw1[H][I] = Dw;
	  } /* fin for I */
	} /* fin for H */

      /* Copia de la capa oculta a las lneas de contexto */
      for (H = 0; H < nHiddenNodes;  H++)
	Context[H] = Out1[L][H];
      }  /* fin for L (Final del entrenamiento de un patrn) */

    /* Guardamos los resultados finales actuales para el patrn */
    for (J = 0; J < nOutputNodes; J++)
       Outend[P][J] = Out2[L-1][J];

    /* Error cometido por los patrones: sumas de diferencias al cuadrado */
    Error = 0.0;
    for (J = 0; J < nOutputNodes; J++)
      Error += (target[L-1][J] - Out2[L-1][J]) * (target[L-1][J] - Out2[L-1][J]);
    } /* fin for P (Final del entrenamiento de todos los patrones) */

  if (kbhit())
    {
    c = getch();
    if ((c = toupper(c)) == 'E')
      MonitorError++; /* Si se pulsa la tecla E se visualiza el error */
    else
      if (c == ESC)
	break;        /* Si se pulsa la tecla ESC abortamos la ejecucin */
    }/* fin if */

  /* Error promedio sobre todos los patrones */
  Error /= nPatterns * nOutputNodes;

  if (MonitorError || (Q % nReportErrors == 0))  /* Visualizamos el error */
    {
    /* Imprime el valor del error */
    fprintf(stderr,"Iteracin %5ld/%-5ld Error %lf\r",
	    Q, nIterations, Error);
    if (Q % nReportErrors == 0)
      fprintf(fpError, "%ld %lf\n", Q, Error);
    } /* fin if */

  MonitorError = 0;

  /* Terminamos la ejecucin del bucle principal si hemos rebajado la cota del error */
  if (Error < ErrorLevel)
      break;

  /* Cerramos el fichero de patrones */
  fclose(fpPattern);
  } /* fin for Q (final del N de iteraciones) */

/* Imprime el valor del error */
fprintf(stderr,"Iteracin %5ld/%-5ld Error %lf\r",
	Q, nIterations, Error);
if (Q % nReportErrors == 0)
  fprintf(fpError, "%ld %lf\n", Q, Error);

/* Cerramos el fichero de errores */
fclose(fpError);

/* Almacena los pesos finales */
if ((fpWeightsOut = fopen(szWeightsOut,"w")) == NULL)
  {
  fprintf(stderr," %s:  Can't write file %s \n",programe,szWeightsOut);
  exit(1);
  } /* fin if */

for (H = 0;  H < nHiddenNodes; H++)
  for (I = 0;  I <= nInputNodes; I++)
    fprintf(fpWeightsOut, "%lf%c", W1[H][I],
	    I%ITEMS==ITEMS-1 ? '\n':' ');
for (J = 0;  J < nOutputNodes;  J++)
  for (H = 0;  H <= nHiddenNodes;  H++)
    fprintf(fpWeightsOut,  "%lf%c", W2[J][H],
	    J%ITEMS==ITEMS-1 ? '\n':' ');

/* Cerramos el fichero de pesos */
fclose(fpWeightsOut);

/* Escribimos el fichero final de activacin */
for (P = 0; P < nPatterns; P++)
  {
  fprintf(fpResults, "%ld\n", P+1);
  for (J = 0; J < nOutputNodes; J++)
    fprintf(fpResults, " %lf", Outend[P][J]);
  fprintf(fpResults, "\n");
  } /* fin for P */

/* Cerramos el fichero de resultados */
fclose(fpResults);

/* Liberar la memoria dinmica */
FreeMatrix(Out0,   MaxLetters + 2);
FreeMatrix(Out1,   MaxLetters + 2);
FreeMatrix(Out2,   MaxLetters + 2);
FreeMatrix(Outend, MaxLetters + 2);
FreeMatrix(Delta1, MaxLetters + 2);
FreeMatrix(Delta2, MaxLetters + 2);
FreeMatrix(target, MaxLetters + 2);
FreeMatrix(Delw1,  nHiddenNodes);
FreeMatrix(W1,     nHiddenNodes);
FreeMatrix(Delw2,  nOutputNodes);
FreeMatrix(W2,     nOutputNodes);
free(Context);
if (c != ESC)
  printf("\nLa ejecucin en Batch %ld termin correctamente.", R+1);
else
  printf("\nLa ejecucin en Batch %ld fue interrumpida por el usuario.", R+1);

} /* fin for R */

/* Cerramos el fichero de ejecucin */
fclose(fpRun);

} /* fin del MAIN */


/* Array storage allocation riutines */
void AllocateVector(VECTOR *Vector, long nCols)
  {
  if ((*Vector = (VECTOR) calloc(nCols, sizeof(double))) == NULL)
    {
    fprintf(stderr, " Sorrt Not enough memory for nodes !\n");
    exit(1);
    }
  }

/* Allocate space for dynamic, two-dimensii */
void AllocateColumns(PFLOAT Matrix[], long nRows, long nCols)
  {
  long i;
  for (i = 0; i < nRows; i++)
    AllocateVector(&Matrix[i], nCols);
  }

void AllocateMatrix(MATRIX *Pmatrix,long nRows, long nCols)
  {
  if ((*Pmatrix = (MATRIX) calloc(nRows, sizeof(PFLOAT))) == NULL)
    {
    fprintf(stderr, " Sorrt Not enough memory for nodes !\n");
    exit(1);
    }
  AllocateColumns(*Pmatrix, nRows, nCols);
  }

/* Free the memory used by the matrix. */
void FreeMatrix(MATRIX Matrix, long nRows)
  {
  long i;
  for (i = 0; i < nRows; i++)
    free(Matrix[i]);
  free(Matrix);
  }

